"""Utils for the LLM Inflation project."""

import numpy as np
import pandas as pd


def log_returns(self, n_shifts=1):
    """Log returns of a time series."""
    return np.log(self / self.shift(n_shifts))


pd.Series.log_returns = log_returns
pd.DataFrame.log_returns = log_returns


def lapeyres_aggregation(
    index_values_children, weights_df, hicp, country, parent_node="CP00"
):
    """
    Implements the Lapeyres Aggregation algorithm.

    Lapeyres Aggregation algorithm:
    1) Take the index values of the children of node 'c', and divide each value by the value of the last December of the previous year.
    2) For each children, this value is multiplied by the current-year weight of the children node.
    3) These values are summed among the children to obtain the evolution of the parent node.
    4) Finally, the value is multiplied by the last December value of the previous year of the parent node.

    The algorithm presented here takes a different but ultimately equivalent approach:
    1) Compute the log returns of the index values of the children of node 'c'.
    2) Compute their intra-year cumulative sums.
    3) For each children, this value is multiplied by the current-year weight of the children node.
    4) These values are summed among the children to obtain the evolution of the parent node.
    5) The resulting time-series is exponentiated to get levels.
    6) Finally, the value is scaled by the last December value of the previous year of the parent node.
    """
    # 1) Compute the log returns of the index values of the children of node 'c'
    mom_reconstructed_index_df = index_values_children.log_returns().iloc[1:]

    # 2) Compute their intra-year cumulative sums
    yearly_cumsum = mom_reconstructed_index_df.groupby(
        mom_reconstructed_index_df.index.year
    ).cumsum()

    # 3) For each children, multiply by the current-year weight of the children node.
    weighted_cumsum = yearly_cumsum.copy()
    for year in yearly_cumsum.index.year.unique():
        weighted_cumsum.loc[weighted_cumsum.index.year == year] *= (
            weights_df.loc[year].values / 1000
        )

    # 4 & 5) Sum the values among the children and exponentiate to get levels
    final_result = np.exp(weighted_cumsum.sum(axis=1))

    # 6) Scale by the last December value of the previous year of the parent node.
    for i, year in enumerate(final_result.index.year.unique()):
        # For the first year, we have to use known values from the previous year
        if i == 0:
            final_result.loc[final_result.index.year == year] *= hicp.prices[country][
                parent_node
            ].loc[f"{year-1}-12-01"]
        # After, we can use the values we just computed
        else:
            final_result.loc[final_result.index.year == year] *= final_result.loc[
                final_result.index.year == year - 1
            ].iloc[-1]

    return final_result



"""Multivariate model for inflation forecasting."""

import pandas as pd
from typing import Union, Optional
from darts import TimeSeries
from darts.models import XGBModel, LinearRegressionModel, RandomForest
from darts.metrics import mae

multivariate_models = [
    {
        "name": "LinearRegression",
        "model": LinearRegressionModel,
        "kwargs": {},
    },
    {
        "name": "RandomForest",
        "model": RandomForest,
        "kwargs": {
            "n_estimators": 100,
            "max_depth": 10,
        },
    },
    {
        "name": "XGBoost",
        "model": XGBModel,
        "kwargs": {},
    },
]


class MultivariateForecaster:
    """Multivariate model for inflation forecasting."""

    def __init__(
        self,
        model_type: str,
        horizon: int,
        lags: Union(int, list[int]),
        prices: TimeSeries,
        weights: pd.DataFrame,
        covariates: Optional(TimeSeries) = None,
        lags_covariates: Optional(Union(int, list[int])) = None,
    ):
        """
        Initialize the model.

        Parameters
        ----------
        model_type : str
            The type of model to use.
            ["LinearRegression", "RandomForest", "XGBoost"]
        horizon : int
            The number of periods to forecast.
        lags : int or list
            The lags to use for the model (positive int or negative list)
        prices : TimeSeries
            The prices to forecast.
        weights : pd.DataFrame
            The weights to use for aggregation.
        covariates : TimeSeries (optional)
            The covariates to use for the model.
        lags_covariates : list (optional)
            The lags to use for the covariates.
        """
        self.prices = prices
        self.weights = weights
        self.covariates = covariates
        self.lags_covariates = lags_covariates
        self.horizon = horizon
        self.lags = lags
        self.model_type = model_type

        model_params = [
            model for model in multivariate_models if model["name"] == model_type
        ][0]
        self.model_class = model_params["model"]
        self.model_kwargs = model_params["kwargs"]

        self._init_model()

    def _init_model(self):
        """Initialize the 'darts' model."""
        self.model = self.model_class(
            **dict(
                **{"lags": self.lags, "output_chunk_length": self.horizon},
                **(self.model_kwargs or {}),
                **(
                    {"lags_past_covariates": self.lags_covariates}
                    if self.lags_covariates
                    else {}
                )
            )
        )

    def fit(
        self,
        new_prices: Optional(TimeSeries) = None,
        new_covariates: Optional(TimeSeries) = None,
    ):
        """
        Fit the model. Uses the initial prices and covariates by default.

        Parameters
        ----------
        new_prices : TimeSeries (optional)
            The new prices to use for fitting.
        new_covariates : TimeSeries (optional)
            The new covariates to use for fitting.
        """
        if new_prices is None:
            new_prices = self.prices
        if new_covariates is None:
            new_covariates = self.covariates
        self.model.fit(new_prices, past_covariates=new_covariates)

    def backtest(self, stride: int, metric: callable = mae):
        """
        Run a backtest on the model.

        Parameters
        ----------
        stride : int
            How many steps to move the test window at each iteration.
        metric : callable (optional)
            The metric to use for scoring. Defaults to Mean Absolute Error.
        """
        score = self.model.backtest(
            series=self.prices,
            past_covariates=self.covariates,
            stride=stride,
            metric=metric,
            verbose=True,
        )
        return score

    # def show_predictions2(self, n=6, past_values=12):
    #     self.model.fit(self.prices, past_covariates=self.covariates)
    #     prediction = self.model.predict(n)
    #     prediction.plot(label="prediction")
    #     self.prices[-past_values:].plot(label="actual")
    #     plt.legend()
    #     plt.show()

    # def show_predictions(self, n_predicted=6, past_values=12, end_date=None):
    #     # Ensure end_date is a pd.Timestamp for consistency
    #     if end_date:
    #         end_date = pd.Timestamp(end_date)

    #     # Fit the model up to end_date if specified, else use full data
    #     if end_date:
    #         prices_to_fit = self.prices.drop_after(end_date)
    #         covariates_to_fit = (
    #             self.covariates.drop_after(end_date) if self.covariates else None
    #         )
    #     else:
    #         prices_to_fit = self.prices
    #         covariates_to_fit = self.covariates

    #     # Fit the model
    #     self.model.fit(series=prices_to_fit, past_covariates=covariates_to_fit)

    #     # Generate predictions from the end of the training data
    #     prediction = self.model.predict(n=n_predicted)

    #     # Calculate the start date for the actual past values
    #     start_past = (
    #         end_date - relativedelta(months=past_values)
    #         if end_date
    #         else self.prices.end_time() - relativedelta(months=past_values)
    #     )

    #     # Calculate the end date for actual future values to be included
    #     end_future = (
    #         end_date + relativedelta(months=n_predicted)
    #         if end_date
    #         else self.prices.end_time()
    #     )

    #     # Select the actual values range to include both past and the available future
    #     actual = self.prices[start_past:end_future]

    #     # Plot actual values (both past and future when available)
    #     actual.plot(label="Actual")

    #     # Plot predictions
    #     prediction.plot(label="Predictions")

    #     plt.legend()
    #     plt.title("Predictions vs Actual Values")
    #     plt.xlabel("Time")
    #     plt.ylabel("Value")
    #     plt.show()

    def make_training_data(self, forecast_start: str, prices_df: pd.DataFrame):
        """
        Make training data for the model.

        Parameters
        ----------
        forecast_start : str
            The date from which to start forecasting.
        prices_df : pd.DataFrame
            The DataFrame containing the prices, from cpilib.
        """
        training_data = self.prices.drop_after(pd.Timestamp(forecast_start))
        training_covariates = (
            self.covariates.drop_after(pd.Timestamp(forecast_start))
            if self.covariates
            else None
        )

        true_index = prices_df[prices_df["date"] < forecast_start].set_index("date")

        last_year = true_index.index.year[-1] - 1
        year_before_last = true_index.index.year.unique()[-2] - 1

        # Keep only last year and the december before
        true_index = true_index[
            (
                (true_index.index.year >= last_year)
                | (
                    (true_index.index.year == year_before_last)
                    & (true_index.index.month == 12)
                )
            )
        ]

        return training_data, training_covariates, true_index



"""Functions to collect sentiment scores for each category from the OpenAI API."""

from ecb_scraper import load_ecb_conferences
import pandas as pd
from tqdm import tqdm
from openai import OpenAI
from tenacity import retry, stop_after_attempt, wait_random_exponential

model = "gpt-3.5-turbo-1106"
categories = [
    "Interest rates",
    "Employment",
    "Energy prices",
    "Inflation expectations",
    "GDP growth",
    "Supply Chain Conditions",
    "Wage Growth",
]


def system_(date, categories):
    """System message to reset the context and prompt the user for the next input."""
    return f"""
        Reset context. Analyze the following ECB press conference from {date} without using any knowledge beyond that date.
        Provide a list of sentiment scores for each of the specified categories using a scale from 0 to 9,
        where 0 indicates a future decrease, 4 or 5 indicates no change, and 9 indicates a future increase.
        Categories: {", ".join(categories)}
        The format for the output should be exactly 7 integers, comma-separated, corresponding to each category in the order listed above.
        An example output might look like this: 5,4,4,9,0,4,7
        Please adhere to this format strictly.
    """


@retry(wait=wait_random_exponential(min=1, max=60), stop=stop_after_attempt(6))
def chat_completion(client, user, date, categories, model=model):
    """Send a chat completion request to the OpenAI API."""
    try:
        response = client.chat.completions.create(
            model=model,
            messages=[
                {"role": "system", "content": system_(date, categories)},
                {"role": "user", "content": user},
            ],
            temperature=0,
        )
        return response
    except Exception as e:
        print(f"Error for {date}: {e}")
        return None


def parse_response(text):
    """
    Parse API response.

    Despite the clear instructions, the user might not adhere to the format.
    This function attempts to parse the response.
    GPT4 was much better at following instructions.
    """
    if text.startswith("Interest rates"):
        scores = [int(str_[0]) for str_ in text.split(": ")[1:]]
    else:
        scores = [int(str_[0]) for str_ in text.replace(" ", "").split(",")]
    return scores


def main(api_key_path="key", output_file="output.csv"):
    """Main function to collect sentiment scores for each category from the OpenAI API."""
    # Load OpenAI API key
    with open(api_key_path, "r") as f:
        api_key = f.read()
    client = OpenAI(api_key=api_key)

    # Load ECB conferences
    conferences = load_ecb_conferences()

    # Collect responses in a list
    responses = []
    for idx, row in tqdm(conferences.iterrows()):
        response = chat_completion(
            client, user=row["text"], date=row["date"], categories=categories
        )
        responses.append((idx, response))

    raw_output_data = [
        (idx, parse_response(r.choices[0].message.content), conferences["date"][idx])
        for (idx, r) in responses
    ]

    # Convert the raw data into a DataFrame
    df = pd.DataFrame(raw_output_data, columns=["Index", "Scores", "Date"])

    # Split the "Scores" list into separate columns
    for idx, category in enumerate(categories):
        df[category] = df["Scores"].apply(lambda x: x[idx])  # noqa

    df = df.drop(columns=["Scores", "Index"])

    df["Date"] = pd.to_datetime(df["Date"])
    df = df.sort_values(by="Date")

    df.to_csv(output_file, index=False)
